import contextlib
import distutils
import distutils.sysconfig
import io
import logging
import os
import shutil
import tempfile

import setuptools
import setuptools.command.build_ext

import cppimport
from cppimport.filepaths import make_absolute

logger = logging.getLogger(__name__)


def build_module(module_data):
    _handle_strict_prototypes()

    build_path = tempfile.mkdtemp()

    full_module_name = module_data["fullname"]
    filepath = module_data["filepath"]
    cfg = module_data["cfg"]

    module_data["abs_include_dirs"] = [
        make_absolute(module_data["filedirname"], d)
        for d in cfg.get("include_dirs", [])
    ] + [os.path.dirname(filepath)]
    module_data["abs_library_dirs"] = [
        make_absolute(module_data["filedirname"], d)
        for d in cfg.get("library_dirs", [])
    ]
    module_data["dependency_dirs"] = module_data["abs_include_dirs"] + [
        module_data["filedirname"]
    ]
    module_data["extra_source_filepaths"] = [
        make_absolute(module_data["filedirname"], s) for s in cfg.get("sources", [])
    ]

    ext = ImportCppExt(
        os.path.dirname(filepath),
        full_module_name,
        language="c++",
        sources=(
            module_data["extra_source_filepaths"]
            + [module_data["rendered_src_filepath"]]
        ),
        include_dirs=module_data["abs_include_dirs"],
        extra_compile_args=cfg.get("extra_compile_args", []),
        extra_link_args=cfg.get("extra_link_args", []),
        library_dirs=module_data["abs_library_dirs"],
        libraries=cfg.get("libraries", []),
    )

    args = [
        "build_ext",
        "--inplace",
        "--build-temp=" + build_path,
        "--build-lib=" + build_path,
        "-v",
    ]

    setuptools_args = dict(
        name=full_module_name,
        ext_modules=[ext],
        script_args=args,
        cmdclass={"build_ext": BuildImportCppExt},
    )

    # Monkey patch in the parallel compiler if requested.
    # TODO: this will still cause problems if there is multithreaded code
    # interacting with distutils. Ideally, we'd just subclass CCompiler
    # instead.
    if cfg.get("parallel"):
        old_compile = distutils.ccompiler.CCompiler.compile
        distutils.ccompiler.CCompiler.compile = _parallel_compile

    f = io.StringIO()
    with contextlib.redirect_stdout(f):
        with contextlib.redirect_stderr(f):
            setuptools.setup(**setuptools_args)
    logger.debug(f"Setuptools/compiler output: {f.getvalue()}")

    # Remove the parallel compiler to not corrupt the outside environment.
    if cfg.get("parallel"):
        distutils.ccompiler.CCompiler.compile = old_compile

    shutil.rmtree(build_path)


def _handle_strict_prototypes():
    if not cppimport.settings["remove_strict_prototypes"]:
        return

    cfg_vars = distutils.sysconfig.get_config_vars()
    for key, value in cfg_vars.items():
        if type(value) == str:
            cfg_vars[key] = value.replace("-Wstrict-prototypes", "")


class ImportCppExt(setuptools.Extension):
    """
    Subclass setuptools.Extension to add self.libdest specifying where the shared
    library should be placed after being compiled with BuildImportCppExt.
    """

    def __init__(self, libdest, *args, **kwargs):
        self.libdest = libdest
        setuptools.Extension.__init__(self, *args, **kwargs)


class BuildImportCppExt(setuptools.command.build_ext.build_ext):
    """
    Subclass setuptools build_ext to put the compiled shared library in the
    appropriate place in the source tree from the ImportCppExt.libdest value.
    """

    def copy_extensions_to_source(self):
        for ext in self.extensions:
            fullname = self.get_ext_fullname(ext.name)
            filename = self.get_ext_filename(fullname)
            src_filename = os.path.join(self.build_lib, filename)
            dest_filename = os.path.join(ext.libdest, os.path.basename(filename))

            distutils.file_util.copy_file(
                src_filename, dest_filename, verbose=self.verbose, dry_run=self.dry_run
            )


# Patch for parallel compilation with distutils
# From: http://stackoverflow.com/questions/11013851/speeding-up-build-process-with-distutils # noqa: E501
def _parallel_compile(
    self,
    sources,
    output_dir=None,
    macros=None,
    include_dirs=None,
    debug=0,
    extra_preargs=None,
    extra_postargs=None,
    depends=None,
):

    # these lines are copied directly from distutils.ccompiler.CCompiler
    macros, objects, extra_postargs, pp_opts, build = self._setup_compile(
        output_dir, macros, include_dirs, sources, depends, extra_postargs
    )
    cc_args = self._get_cc_args(pp_opts, debug, extra_preargs)

    # Determine the number of compilation threads. Unless there are special
    # circumstances, this is the number of cores on the machine
    N = 1
    try:
        import multiprocessing
        import multiprocessing.pool

        N = multiprocessing.cpu_count()
    except (ImportError, NotImplementedError):
        pass

    def _single_compile(obj):
        try:
            src, ext = build[obj]
        except KeyError:
            return
        self._compile(obj, src, ext, cc_args, extra_postargs, pp_opts)

    # imap is evaluated on demand, converting to list() forces execution
    list(multiprocessing.pool.ThreadPool(N).imap(_single_compile, objects))
    return objects



import argparse
import logging
import os
import sys

from cppimport import build_all, build_filepath, settings


def _run_from_commandline(raw_args):
    parser = argparse.ArgumentParser("cppimport")

    parser.add_argument(
        "--verbose", "-v", action="store_true", help="Increase log verbosity."
    )
    parser.add_argument(
        "--quiet", "-q", action="store_true", help="Only print critical log messages."
    )

    subparsers = parser.add_subparsers(dest="action", required=True)

    build_parser = subparsers.add_parser(
        "build",
        help="Build one or more cpp source files.",
    )
    build_parser.add_argument(
        "root",
        help="The file or directory to build. If a directory is given, "
        "cppimport walks it recursively to build all eligible source "
        "files.",
        nargs="*",
    )
    build_parser.add_argument(
        "--force", "-f", action="store_true", help="Force rebuild."
    )

    args = parser.parse_args(raw_args[1:])

    if args.quiet:
        logging.basicConfig(level=logging.CRITICAL)
    elif args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    if args.action == "build":
        if args.force:
            settings["force_rebuild"] = True

        for path in args.root or ["."]:
            path = os.path.abspath(os.path.expandvars(path))
            if os.path.isfile(path):
                build_filepath(path)
            elif os.path.isdir(path):
                build_all(path or os.getcwd())
            else:
                raise FileNotFoundError(
                    f'The given root path "{path}" could not be found.'
                )
    else:
        parser.print_usage()


if __name__ == "__main__":
    _run_from_commandline(sys.argv)



import logging
import sys
import traceback

import cppimport

logger = logging.getLogger(__name__)


class Hook(object):
    def __init__(self):
        self._running = False

    def find_spec(self, fullname, path, target=None):
        # Prevent re-entry by the underlying importer
        if self._running:
            return

        try:
            self._running = True
            cppimport.imp(fullname, opt_in=True)
        except ImportError:
            # ImportError should be quashed because that simply means cppimport
            # didn't find anything, and probably shouldn't have found anything!
            logger.debug(traceback.format_exc())
        finally:
            self._running = False


# Add the hook to the list of import handlers for Python.
hook_obj = Hook()
sys.meta_path.insert(0, hook_obj)



"""
See CONTRIBUTING.md for a description of the project structure and the internal logic.
"""
import ctypes
import logging
import os

from cppimport.find import _check_first_line_contains_cppimport

try:
    from ._version import version as __version__
    from ._version import version_tuple
except ImportError:
    __version__ = "unknown version"
    version_tuple = (0, 0, "unknown version")

settings = dict(
    force_rebuild=False,  # `force_rebuild` with multiple processes is not supported
    file_exts=[".cpp", ".c"],
    rtld_flags=ctypes.RTLD_LOCAL,
    lock_suffix=".lock",
    lock_timeout=10 * 60,
    remove_strict_prototypes=True,
    release_mode=os.getenv("CPPIMPORT_RELEASE_MODE", "0").lower()
    in ("true", "yes", "1"),
)
_logger = logging.getLogger("cppimport")


def imp(fullname, opt_in=False):
    """
    `imp` is the explicit alternative to using cppimport.import_hook.

    Parameters
    ----------
    fullname : the name of the module to import.
    opt_in : should we require C++ files to opt in via adding "cppimport" to
             the first line of the file? This is on by default for the
             import hook, but is off by default for this function since the
             intent to import a C++ module is clearly specified.

    Returns
    -------
    module : the compiled and loaded Python extension module
    """
    from cppimport.find import find_module_cpppath

    # Search through sys.path to find a file that matches the module
    filepath = find_module_cpppath(fullname, opt_in)
    return imp_from_filepath(filepath, fullname)


def imp_from_filepath(filepath, fullname=None):
    """
    `imp_from_filepath` serves the same purpose as `imp` except allows
    specifying the exact filepath of the C++ file.

    Parameters
    ----------
    filepath : the filepath to the C++ file to build and import.
    fullname : the name of the module to import. This can be different from the
               module name inferred from the filepath if desired.

    Returns
    -------
    module : the compiled and loaded Python extension module
    """
    from cppimport.importer import (
        build_safely,
        is_build_needed,
        load_module,
        setup_module_data,
        try_load,
    )

    filepath = os.path.abspath(filepath)
    if fullname is None:
        fullname = os.path.splitext(os.path.basename(filepath))[0]
    module_data = setup_module_data(fullname, filepath)
    # The call to try_load is necessary here because there are times when the
    # only evidence a rebuild is needed comes from attempting to load an
    # existing extension module. For example, if the extension was built on a
    # different architecture or with different Python headers and will produce
    # an error when loaded, then the load will fail. In that situation, we will
    # need to rebuild.
    if is_build_needed(module_data) or not try_load(module_data):
        build_safely(filepath, module_data)
    load_module(module_data)
    return module_data["module"]


def build(fullname):
    """
    `build` builds a extension module like `imp` but does not import the
    extension.

    Parameters
    ----------
    fullname : the name of the module to import.

    Returns
    -------
    ext_path : the path to the compiled extension.
    """
    from cppimport.find import find_module_cpppath

    # Search through sys.path to find a file that matches the module
    filepath = find_module_cpppath(fullname)
    return build_filepath(filepath, fullname=fullname)


def build_filepath(filepath, fullname=None):
    """
    `build_filepath` builds a extension module like `build` but allows
    to directly specify a file path.

    Parameters
    ----------
    filepath : the filepath to the C++ file to build.
    fullname : the name of the module to build.

    Returns
    -------
    ext_path : the path to the compiled extension.
    """
    from cppimport.importer import (
        build_safely,
        is_build_needed,
        load_module,
        setup_module_data,
    )

    filepath = os.path.abspath(filepath)
    if fullname is None:
        fullname = os.path.splitext(os.path.basename(filepath))[0]
    module_data = setup_module_data(fullname, filepath)
    if is_build_needed(module_data):
        build_safely(filepath, module_data)
    load_module(module_data)
    # Return the path to the built module
    return module_data["ext_path"]


def build_all(root_directory):
    """
    `build_all` builds a extension module like `build` for each eligible (that is,
    containing the "cppimport" header) source file within the given `root_directory`.

    Parameters
    ----------
    root_directory : the root directory to search for cpp source files in.
    """
    for directory, _, files in os.walk(root_directory):
        for file in files:
            if (
                not file.startswith(".")
                and os.path.splitext(file)[1] in settings["file_exts"]
            ):
                full_path = os.path.join(directory, file)
                if _check_first_line_contains_cppimport(full_path):
                    _logger.info(f"Building: {full_path}")
                    build_filepath(full_path)


######## BACKWARDS COMPATIBILITY #########
# Below here, we pay penance for mistakes.
# TODO: Add DeprecationWarning

"""
For backwards compatibility, support this alias for the imp function
"""
cppimport = imp


def force_rebuild(to=True):
    settings["force_rebuild"] = to


def turn_off_strict_prototypes():
    pass  # turned off by default.


def set_rtld_flags(flags):
    settings["rtld_flags"] = flags



import contextlib
import copy
import logging
import os
import shutil
import subprocess
import sys
from multiprocessing import Process
from tempfile import TemporaryDirectory

import cppimport
import cppimport.build_module
import cppimport.templating
from cppimport.find import find_module_cpppath

root_logger = logging.getLogger()
root_logger.setLevel(logging.DEBUG)

handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.DEBUG)
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
handler.setFormatter(formatter)
root_logger.addHandler(handler)


@contextlib.contextmanager
def appended(filename, text):
    with open(filename, "r") as f:
        orig = f.read()
    with open(filename, "a") as f:
        f.write(text)
    try:
        yield
    finally:
        with open(filename, "w") as f:
            f.write(orig)


def subprocess_check(test_code, returncode=0):
    p = subprocess.run(
        ["python", "-c", test_code],
        cwd=os.path.dirname(__file__),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    if len(p.stdout) > 0:
        print(p.stdout.decode("utf-8"))
    if len(p.stderr) > 0:
        print(p.stderr.decode("utf-8"))
    assert p.returncode == returncode


@contextlib.contextmanager
def tmp_dir(files=None):
    """Create a temporary directory and copy `files` into it. `files` can also
    include directories."""
    files = files if files else []

    with TemporaryDirectory() as tmp_path:
        for f in files:
            if os.path.isdir(f):
                shutil.copytree(f, os.path.join(tmp_path, os.path.basename(f)))
            else:
                shutil.copyfile(f, os.path.join(tmp_path, os.path.basename(f)))
        yield tmp_path


def test_find_module_cpppath():
    mymodule_loc = find_module_cpppath("mymodule")
    mymodule_dir = os.path.dirname(mymodule_loc)
    assert os.path.basename(mymodule_loc) == "mymodule.cpp"

    apackage = find_module_cpppath("apackage.mymodule")
    apackage_correct = os.path.join(mymodule_dir, "apackage", "mymodule.cpp")
    assert apackage == apackage_correct

    inner = find_module_cpppath("apackage.inner.mymodule")
    inner_correct = os.path.join(mymodule_dir, "apackage", "inner", "mymodule.cpp")
    assert inner == inner_correct


def test_get_rendered_source_filepath():
    rendered_path = cppimport.templating.get_rendered_source_filepath("abc.cpp")
    assert rendered_path == ".rendered.abc.cpp"


def module_tester(mod, cheer=False):
    assert mod.add(1, 2) == 3
    if cheer:
        mod.Thing().cheer()


def test_mymodule():
    mymodule = cppimport.imp("mymodule")
    module_tester(mymodule)


def test_mymodule_build():
    cppimport.build("mymodule")
    import mymodule

    module_tester(mymodule)


def test_mymodule_from_filepath():
    mymodule = cppimport.imp_from_filepath("tests/mymodule.cpp")
    module_tester(mymodule)


def test_package_mymodule():
    mymodule = cppimport.imp("apackage.mymodule")
    module_tester(mymodule)


def test_inner_package_mymodule():
    mymodule = cppimport.imp("apackage.inner.mymodule")
    module_tester(mymodule)


def test_with_file_in_syspath():
    orig_sys_path = copy.copy(sys.path)
    sys.path.append(os.path.join(os.path.dirname(__file__), "mymodule.cpp"))
    cppimport.imp("mymodule")
    sys.path = orig_sys_path


def test_rebuild_after_failed_compile():
    cppimport.imp("mymodule")
    test_code = """
import cppimport; mymodule = cppimport.imp("mymodule");assert(mymodule.add(1,2) == 3)
"""
    with appended("tests/mymodule.cpp", ";asdf;"):
        subprocess_check(test_code, 1)
    subprocess_check(test_code, 0)


add_to_thing = """
#include <iostream>
struct Thing {
    void cheer() {
        std::cout << "WAHHOOOO" << std::endl;
    }
};
#define THING_DEFINED
"""


def test_no_rebuild_if_no_deps_change():
    cppimport.imp("mymodule")
    test_code = """
import cppimport;
mymodule = cppimport.imp("mymodule");
assert(not hasattr(mymodule, 'Thing'))
"""
    with appended("tests/thing2.h", add_to_thing):
        subprocess_check(test_code)


def test_rebuild_header_after_change():
    cppimport.imp("mymodule")
    test_code = """
import cppimport;
mymodule = cppimport.imp("mymodule");
mymodule.Thing().cheer()
"""
    with appended("tests/thing.h", add_to_thing):
        subprocess_check(test_code)
    assert open("tests/thing.h", "r").read() == ""


def test_raw_extensions():
    raw_extension = cppimport.imp("raw_extension")
    assert raw_extension.add(1, 2) == 3


def test_extra_sources_and_parallel():
    cppimport.settings["force_rebuild"] = True
    mod = cppimport.imp("extra_sources")
    cppimport.settings["force_rebuild"] = False
    assert mod.square_sum(3, 4) == 25


def test_import_hook():
    import cppimport.import_hook

    # Force rebuild to make sure we're not just reloading the already compiled
    # module from disk
    cppimport.force_rebuild(True)
    import hook_test

    cppimport.force_rebuild(False)
    assert hook_test.sub(3, 1) == 2


def test_submodule_import_hook():
    import cppimport.import_hook

    # Force rebuild to make sure we're not just reloading the already compiled
    # module from disk
    cppimport.force_rebuild(True)
    import apackage.mymodule

    cppimport.force_rebuild(False)
    assert apackage.mymodule.add(3, 1) == 4


def test_relative_import():
    import cppimport.import_hook

    cppimport.force_rebuild(True)
    from apackage.rel_import_tester import f

    cppimport.force_rebuild(False)
    print(f())
    assert f() == 3


def test_multiple_processes():
    with tmp_dir(["tests/hook_test.cpp"]) as tmp_path:
        test_code = f"""
import os;
os.chdir('{tmp_path}');
import cppimport.import_hook;
import hook_test;
        """
        processes = [
            Process(target=subprocess_check, args=(test_code,)) for i in range(100)
        ]

        for p in processes:
            p.start()

        for p in processes:
            p.join()

        assert all(p.exitcode == 0 for p in processes)



import importlib
import logging
import os
import sys
import sysconfig
from contextlib import suppress
from time import sleep, time

import filelock

import cppimport
from cppimport.build_module import build_module
from cppimport.checksum import checksum_save, is_checksum_valid
from cppimport.templating import run_templating

logger = logging.getLogger(__name__)


def build_safely(filepath, module_data):
    """Protect against race conditions when multiple processes executing
    `template_and_build`"""
    binary_path = module_data["ext_path"]
    lock_path = binary_path + cppimport.settings["lock_suffix"]

    def build_completed():
        return os.path.exists(binary_path) and is_checksum_valid(module_data)

    t = time()

    # Race to obtain the lock and build. Other processes can wait
    while not build_completed() and time() - t < cppimport.settings["lock_timeout"]:
        try:
            with filelock.FileLock(lock_path, timeout=1):
                if build_completed():
                    break
                template_and_build(filepath, module_data)
        except filelock.Timeout:
            logging.debug(f"Could not obtain lock (pid {os.getpid()})")
            if cppimport.settings["force_rebuild"]:
                raise ValueError(
                    "force_build must be False to build concurrently."
                    "This process failed to claim a filelock indicating that"
                    " a concurrent build is in progress"
                )
            sleep(1)

    if os.path.exists(lock_path):
        with suppress(OSError):
            os.remove(lock_path)

    if not build_completed():
        raise Exception(
            f"Could not compile binary as lock already taken and timed out."
            f" Try increasing the timeout setting if "
            f"the build time is longer (pid {os.getpid()})."
        )


def template_and_build(filepath, module_data):
    logger.debug(f"Compiling {filepath}.")
    run_templating(module_data)
    build_module(module_data)
    checksum_save(module_data)


def setup_module_data(fullname, filepath):
    module_data = dict()
    module_data["fullname"] = fullname
    module_data["filepath"] = filepath
    module_data["filedirname"] = os.path.dirname(module_data["filepath"])
    module_data["filebasename"] = os.path.basename(module_data["filepath"])
    module_data["ext_name"] = get_module_name(fullname) + get_extension_suffix()
    module_data["ext_path"] = os.path.join(
        os.path.dirname(filepath), module_data["ext_name"]
    )
    return module_data


def get_module_name(full_module_name):
    return full_module_name.split(".")[-1]


def get_extension_suffix():
    ext_suffix = sysconfig.get_config_var("EXT_SUFFIX")
    if ext_suffix is None:
        ext_suffix = sysconfig.get_config_var("SO")
    return ext_suffix


def _actually_load_module(module_data):
    module_data["module"] = importlib.import_module(module_data["fullname"])


def load_module(module_data):
    if hasattr(sys, "getdlopenflags"):
        # It can be useful to set rtld_flags to RTLD_GLOBAL. This allows
        # extensions that are loaded later to share the symbols from this
        # extension. This is primarily useful in a project where several
        # interdependent extensions are loaded but it's undesirable to combine
        # the multiple extensions into a single extension.
        old_flags = sys.getdlopenflags()
        new_flags = old_flags | cppimport.settings["rtld_flags"]
        sys.setdlopenflags(new_flags)
        _actually_load_module(module_data)
        sys.setdlopenflags(old_flags)
    else:
        _actually_load_module(module_data)


def is_build_needed(module_data):
    if cppimport.settings["force_rebuild"]:
        return True
    if cppimport.settings["release_mode"]:
        logger.debug(
            f"Release mode is enabled. Thus, file {module_data['filepath']} is "
            f"not being compiled."
        )
        return False
    if not is_checksum_valid(module_data):
        return True
    logger.debug(f"Matching checksum for {module_data['filepath']} --> not compiling")
    return False


def try_load(module_data):
    """Try loading the module to test if it's not corrupt and for the correct
    architecture"""
    try:
        load_module(module_data)
        return True
    except ImportError as e:
        logger.info(
            f"ImportError during import with matching checksum: {e}. Trying to rebuild."
        )
        with suppress(OSError):
            os.remove(module_data["fullname"])
        return False
